/*
 * Copyright(C) 2010 Canonical Ltd.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Ken VaDine <ken.vandine@canonical.com>
 */

class TelepathyIndicator : GLib.Object {

	public MainLoop loop;
	private TelepathyGLib.DBusDaemon tp_dbus_daemon;
	private static GLib.HashTable<TelepathyGLib.Channel,GLib.HashTable> indicators;
	private static Indicate.Server server;
	private GLib.HashTable<string,GLib.Value?> channel_table;
	private static TelepathyGLib.Channel channel;

	public TelepathyIndicator ()
	{
		indicators = new GLib.HashTable<TelepathyGLib.Channel,GLib.HashTable>(direct_hash, direct_equal);
		server = Indicate.Server.ref_default();
		server.set_type("message.im");
		var desktop_file = GLib.Path.build_filename(
			GLib.Path.DIR_SEPARATOR_S, "usr", "share", 
			"applications", "telepathy-indicator.desktop");
		if (GLib.FileUtils.test(GLib.Path.build_filename(GLib.Path.DIR_SEPARATOR_S, GLib.Environment.get_current_dir(), "data", "telepathy-indicator.desktop"), GLib.FileTest.EXISTS)) {
			desktop_file = GLib.Path.build_filename(GLib.Path.DIR_SEPARATOR_S, GLib.Environment.get_current_dir(), "data", "telepathy-indicator.desktop");
		} else if (GLib.FileUtils.test(GLib.Path.build_filename(GLib.Path.DIR_SEPARATOR_S, "usr", "local", "share", "applications", "telepathy-indicator.desktop"), GLib.FileTest.EXISTS)) {
			desktop_file = GLib.Path.build_filename(GLib.Path.DIR_SEPARATOR_S, "usr", "local", "share", "applications", "telepathy-indicator.desktop");
		}
		server.set_desktop_file(desktop_file);
		server.server_display.connect(server_display);
		server.show();
	}
	void run()
	{
		this.loop = new MainLoop ();

		if (GLib.Environment.get_variable("TELEPATHY_INDICATOR_DEBUG") == "1") {
			debug ("Running in DEBUG mode");
			TelepathyGLib.debug_set_flags("all");
		}

		try {
			tp_dbus_daemon = TelepathyGLib.DBusDaemon.dup();
	        } catch (GLib.Error e) {
			error ("Error: %s", e.message);
		}
	
		var text_filter = new GLib.HashTable<weak string?,weak GLib.Value?>(str_hash, str_equal);
		/* Private text channels */
		text_filter.insert(TelepathyGLib.PROP_CHANNEL_CHANNEL_TYPE, TelepathyGLib.IFACE_CHANNEL_TYPE_TEXT);
		/* File transfer */
		var file_transfer_filter = new GLib.HashTable<weak string?,weak GLib.Value?>(str_hash, str_equal);
		file_transfer_filter.insert(TelepathyGLib.PROP_CHANNEL_CHANNEL_TYPE, TelepathyGLib.IFACE_CHANNEL_TYPE_FILE_TRANSFER);
		/* Calls */
		var call_filter = new GLib.HashTable<weak string?,weak GLib.Value?>(str_hash, str_equal);
		call_filter.insert(TelepathyGLib.PROP_CHANNEL_CHANNEL_TYPE, TelepathyGLib.IFACE_CHANNEL_TYPE_STREAMED_MEDIA);
		var approver = new TelepathyGLib.SimpleApprover(tp_dbus_daemon , "IndicatorApprover", true, (TelepathyGLib.SimpleApproverAddDispatchOperationImpl) add_dispatch_operation_cb, null, (GLib.DestroyNotify) tpi_dispose);

		approver.add_approver_filter(text_filter);
		approver.add_approver_filter(file_transfer_filter);
		approver.add_approver_filter(call_filter);

		var observer = new TelepathyGLib.SimpleObserver(tp_dbus_daemon, false, "IndicatorObserver", true, (TelepathyGLib.SimpleObserverObserveChannelsImpl) add_observer_observe_cb, null, (GLib.DestroyNotify) tpi_dispose);
		observer.add_observer_filter(text_filter);
		observer.add_observer_filter(file_transfer_filter);
		observer.add_observer_filter(call_filter);

		try {
			approver.register();
	        } catch (GLib.Error e) {
			error ("Error: %s", e.message);
		}

		try {
			//observer.register();
	        } catch (GLib.Error e) {
			error ("Error: %s", e.message);
		}

		this.loop.run ();
	}

	void tpi_dispose (void* data)
	{
		debug ("tpi_dispose called");
	}

	void add_indicator(GLib.HashTable indicators, TelepathyGLib.Channel channel, TelepathyGLib.Account account, TelepathyGLib.Connection connection, GLib.List<TelepathyGLib.ChannelRequest> requests, TelepathyGLib.ChannelDispatchOperation? dispatch_operation)
	{
		debug ("add_indicator called");
		this.channel = channel;
		GLib.HashTable<TelepathyGLib.Channel,GLib.HashTable> priv_indicators = indicators;
                uint[] features = {
                        TelepathyGLib.ContactFeature.ALIAS,
                        TelepathyGLib.ContactFeature.PRESENCE,
                        TelepathyGLib.ContactFeature.CONTACT_INFO
                };

                TelepathyGLib.Handle handle;
		var handle_type = TelepathyGLib.HandleType.CONTACT;
                uint[] handles = {};
		handle = this.channel.get_handle(out handle_type);
		handles += handle;

		debug ("there are %d handles", handles.length);
		debug ("there are %u requests", requests.length());

		debug ("object type is %s", this.channel.get_type().name());
		debug ("identifier is %s", this.channel.identifier);
		if(priv_indicators.size() > 0) {
			foreach (var k in priv_indicators.get_keys())
			{
				if (k.get_object_path() == this.channel.get_object_path()) {
					debug ("Found indicator for %s", this.channel.get_object_path());
					return;
				}
			}
		}

		debug ("No indicator found, adding a new one");
		var indicator = new Indicate.Indicator();
		indicator.set_property_value("subtype", "im");
		indicator.set_property_value("sender", this.channel.identifier);
		indicator.set_property_value("body", "Testing");
		var tv = GLib.TimeVal();
		indicator.set_property_time("time", tv);
		indicator.set_property_bool("draw-attention", true);
		indicator.user_display.connect(user_display);
		indicator.show();
		channel_table = new GLib.HashTable<string,GLib.Value?>(str_hash, str_equal);
		channel_table.insert("account", account);
		channel_table.insert("connection", connection);
		channel_table.insert("indicator", indicator);
		if (dispatch_operation != null)
		{
			channel_table.insert("dispatch_operation", dispatch_operation);
		}
		priv_indicators.insert(this.channel, channel_table);

		this.channel.chat_state_changed.connect(channel_chat_state_changed_cb);
		debug ("Chat state is %d", this.channel.get_chat_state(handle));
		this.channel.chat_state_changed.connect((source) => {
			debug ("channel state changed");});
		debug ("before get_contacts_by_handle");
		connection.get_contacts_by_handle(handles, features, (TelepathyGLib.ConnectionContactsByHandleCb) lookup_contact_by_handle_cb, null);  
		debug ("after get_contacts_by_handle");
	}

	void channel_chat_state_changed_cb (uint contact, uint state)
	{
		debug ("channel_chat_state_changed_cb contact is %u", contact);
		debug ("channel_chat_state_changed_cb state is %u", state);
	}

	void add_dispatch_operation_cb (TelepathyGLib.Account account, TelepathyGLib.Connection connection, GLib.List<TelepathyGLib.Channel> channels, TelepathyGLib.ChannelDispatchOperation? dispatch_operation, TelepathyGLib.AddDispatchOperationContext context, void* user_data)
	{
		debug ("add_dispatch_operation_cb");
		foreach (TelepathyGLib.Channel c in channels) {
			this.channel = c;
			debug ("channel = %s", this.channel.get_object_path());
			debug ("Type: %s", this.channel.get_channel_type());
			add_indicator(indicators, this.channel, account, connection, null, dispatch_operation);
		}
		context.accept ();
	}

	void add_observer_observe_cb (TelepathyGLib.Account account, TelepathyGLib.Connection connection, GLib.List<TelepathyGLib.Channel> channels, TelepathyGLib.ChannelDispatchOperation? dispatch_operation, GLib.List<TelepathyGLib.ChannelRequest> requests, TelepathyGLib.ObserveChannelsContext context, void* user_data)
	{
		debug("add_observer_observe_cb");

		var handle_type = TelepathyGLib.HandleType.CONTACT;
		TelepathyGLib.Handle self_handle;
		TelepathyGLib.Handle handle;

		foreach (TelepathyGLib.Channel c in channels) {
			this.channel = c;
			debug ("channel = %s", this.channel.get_object_path());
			debug ("Type: %s", this.channel.get_channel_type());
			self_handle = connection.get_self_handle();
			handle = this.channel.get_handle(out handle_type);
			debug ("self_handle is %d", self_handle);
			debug ("handle is %d", handle);

			if (self_handle == handle)
			{
				debug ("I initiated this channel");
				return;
			};
			if (dispatch_operation == null)
			{
				debug ("dispatch_operation is null");
			}
			add_indicator(indicators, this.channel, account, connection, requests, dispatch_operation);
		}
		context.accept ();
	}

        public void lookup_contact_by_handle_cb (TelepathyGLib.Connection connection, TelepathyGLib.Contact[] contacts, TelepathyGLib.Handle[] failed, GLib.Error error, GLib.Object *weak_object)
	{
		debug ("lookup_contacts_cb");
		GLib.HashTable<string,GLib.Value?> channel_table;

		if (indicators.size() > 0) {
			channel_table = indicators.lookup(channel);
			//debug ("channel is of type %s", channel.get_type().name());
			//debug ("channel is %s", channel.get_object_path());
			var indicator = (Indicate.Indicator) channel_table.lookup("indicator");
			if (indicator is Indicate.Indicator) {
				debug ("found the indicator");
				indicator.set_property_value("sender", contacts[0].get_alias());
			}
		}

		for (int i = 0; i <= contacts.length; i ++)
		{
			var contact = contacts[i];
			debug ("alias is %s", contact.get_alias());
			debug ("identifier is %s", contact.identifier);
		}

		debug ("FAILED is %d", failed.length);
	}

	void server_display()
	{
		debug ("Server activated");
		bool ret = GLib.Process.spawn_command_line_async("empathy");
		if (ret)
		{
			debug ("Contact list raised");
		}
	}
	
	void user_display (Indicate.Indicator indicator, uint timestamp)
	{
		debug ("User activated");
		if (indicator is Indicate.Indicator) {
			if (this.indicators.size() > 0) {
				foreach (var channel in this.indicators.get_keys()) 
				{
					debug ("object type is %s", channel.get_type().name());
					debug ("got a key %s", channel.get_object_path());
					channel_table = this.indicators.lookup(channel);
					if (indicator == channel_table.lookup("indicator")) {
						debug ("Indicator found");
						debug ("CHANNEL is %s", channel.get_object_path());
						var dispatch_operation = channel_table.lookup("dispatch_operation");
						if (dispatch_operation != null)
						{
							debug ("Got dispatch operation");
							var priv_dispatch_operation = (TelepathyGLib.ChannelDispatchOperation) dispatch_operation;
							priv_dispatch_operation.handle_with_time_async("", (int64) Gdk.Event.get_time, (GLib.AsyncReadyCallback) handle_channel_cb);
						} else
						{
							debug ("No dispatch operation, must be from an observer");
							//var account = (TelepathyGLib.Account) channel_table.lookup("account");
							//var request = channel_table.lookup("request");
							//var channel_request = new TelepathyGLib.AccountChannelRequest(account, request, (int64) Gdk.Event.get_time);
						}
						this.indicators.remove(channel);
					}
				}
			}
			indicator.hide();
		}
	}
	public async void handle_channel_cb (GLib.AsyncResult result)
	{
		var dispatch_operation = (TelepathyGLib.ChannelDispatchOperation) result.get_source_object();
		debug ("dispatch_operation is %s", dispatch_operation.get_object_path());
		//if (GLib.SimpleAsyncResult.is_valid(result, result.get_source_object(), null))
		//{
		//	debug ("Got valid result");
		//}
		if (result != null)
		{
			debug ("Got result");
		}
	}

    static int main (string[] args) {
        try {
            var telepathy_indicator = new TelepathyIndicator ();
            telepathy_indicator.run ();
        } catch (GLib.Error e) {
            error ("Error: %s", e.message);
            return 1;
        }
        return 0;
    }
}
